# Copyright (C) 2017-2023 Pier Carlo Chiodi
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
REJECT_REASONS = {
"1": "Invalid AS_PATH length",
"2": "Prefix is bogon",
"3": "Prefix is in global blacklist",
"4": "Invalid AFI",
"5": "Invalid NEXT_HOP",
"6": "Invalid left-most ASN",
"7": "Invalid ASN in AS_PATH",
"8": "Transit-free ASN in AS_PATH",
"9": "Origin ASN not in IRRDB AS-SETs",
"10": "IPv6 prefix not in global unicast space",
"11": "Prefix is in client blacklist",
"12": "Prefix not in IRRDB AS-SETs",
"13": "Invalid prefix length",
"14": "RPKI INVALID route",
"15": "Never via route-servers ASN in AS_PATH",
}
