router id {{ cfg.router_id }};
define rs_as = {{ cfg.rs_as }};

{% if not "logging" is used_local_file %}
log "/var/log/bird.log" all;
log syslog all;
debug protocols { states, routes, filters, interfaces, events };
{% else %}
{{ "logging"|include_local_file -}}
{% endif %}

timeformat base         iso long;
timeformat log          iso long;
timeformat protocol     iso long;
timeformat route        iso long;

protocol device {};

{% if cfg.path_hiding -%}
{%   if "2.0.0"|target_version_le -%}
table master sorted;
{%   else -%}
{%     for this_ip_ver in list_ip_vers %}
ipv{{ this_ip_ver }} table master{{ this_ip_ver }} sorted;
{%     endfor %}
{%   endif -%}
{%- endif %}
